from abc import ABC
from dataclasses import dataclass
from typing import Any

from .base import NeatValidationError, ValidationWarning

__all__ = [
    "ModelImportWarning",
    "UnknownComponentWarning",
    "UnknownSubComponentWarning",
    "IgnoredComponentWarning",
    "UnknownPropertyWarning",
    "UnknownValueTypeWarning",
    "MissingContainerWarning",
    "MissingContainerPropertyWarning",
    "MultipleDataModelsWarning",
    "UnknownPropertyTypeWarning",
    "FailedToInferValueTypeWarning",
    "MoreThanOneNonAlphanumericCharacterWarning",
    "UnknownContainerConstraintWarning",
    "NoDataModelError",
    "ModelImportError",
    "InvalidComponentError",
    "MissingParentDefinitionError",
    "MissingIdentifierError",
    "UnsupportedPropertyTypeError",
    "APIError",
    "FailedImportWarning",
]


@dataclass(frozen=True)
class ModelImportWarning(ValidationWarning, ABC):
    description = "A warning was raised during importing."


@dataclass(frozen=True)
class UnknownComponentWarning(ModelImportWarning):
    description = "Unknown component this will be ignored in the imports."
    fix = "Check if the component is defined in the source file."

    component_type: str
    instance_name: str | None = None
    instance_id: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["component_type"] = self.component_type
        output["instance_name"] = self.instance_name
        output["instance_id"] = self.instance_id
        return output

    def message(self) -> str:
        if self.instance_name:
            prefix = f"Unknown component of type'{self.component_type}' with name '{self.instance_name}'."
        else:
            prefix = f"Unknown component '{self.component_type}'"
        return f"{prefix} This will be ignored in the imports."


@dataclass(frozen=True)
class UnknownSubComponentWarning(UnknownComponentWarning):
    sub_component: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["sub_component"] = self.sub_component
        return output

    def message(self) -> str:
        if self.instance_name:
            prefix = f"Unknown sub-component of type'{self.component_type}' with name '{self.instance_name}'."
        else:
            prefix = f"Unknown sub-component '{self.component_type}'"
        return f"{prefix} This will be ignored in the imports."


@dataclass(frozen=True)
class IgnoredComponentWarning(ModelImportWarning):
    description = "This will be ignored in the imports."
    fix = "No fix is available. "

    reason: str
    identifier: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["reason"] = self.reason
        output["identifier"] = self.identifier
        return output

    def message(self) -> str:
        if self.identifier:
            prefix = f"Identifier '{self.identifier}.' is ignored."
        else:
            prefix = "This is ignored."
        return f"{prefix} {self.reason}"


@dataclass(frozen=True)
class UnknownPropertyWarning(ValidationWarning):
    description = "Unknown property this will be ignored in the imports."
    fix = "Check if the property is defined in the DTDL file."

    component_type: str
    property_name: str
    instance_name: str | None = None
    instance_id: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["component_type"] = self.component_type
        output["property_name"] = self.property_name
        output["instance_name"] = self.instance_name
        output["instance_id"] = self.instance_id
        return output

    def message(self) -> str:
        if self.instance_name:
            prefix = (
                f"Unknown property '{self.property_name}' of component "
                f"'{self.component_type}' with name '{self.instance_name}'."
            )
        else:
            prefix = f"Unknown property '{self.property_name}' of component '{self.component_type}'"
        return f"{prefix} This will be ignored in the imports."


@dataclass(frozen=True)
class UnknownValueTypeWarning(ModelImportWarning):
    description = "Unknown value type. This limits validation done by NEAT. "
    fix = "Set the value type"
    class_id: str
    property_id: str

    def dump(self) -> dict[str, str]:
        return {"class_id": self.class_id, "property_id": self.property_id}

    def message(self) -> str:
        return (
            f"Unknown value type for property '{self.property_id}' of class '{self.class_id}'. "
            "This limits validation done by NEAT."
        )


@dataclass(frozen=True)
class MultipleDataModelsWarning(ModelImportWarning):
    description = "Multiple data models detected. This is not supported."
    fix = "Remove the extra data models."

    data_models: list[str]

    def message(self) -> str:
        return f"Multiple data models detected: {self.data_models}. Will only import the first one."

    def dump(self) -> dict[str, list[str]]:
        return {"data_models": self.data_models}


@dataclass(frozen=True)
class MissingContainerWarning(ModelImportWarning):
    description = "Missing container definition."
    fix = "Add a container definition."

    view_id: str
    property_: str
    container_id: str

    def message(self) -> str:
        return (
            f"Container '{self.container_id}' is missing. "
            f"Will skip property '{self.property_}' of view '{self.view_id}'."
        )

    def dump(self) -> dict[str, str]:
        return {"view_id": self.view_id, "property": self.property_, "container_id": self.container_id}


@dataclass(frozen=True)
class MissingContainerPropertyWarning(ModelImportWarning):
    description = "Missing container property definition."
    fix = "Add a container property definition."

    view_id: str
    property_: str
    container_id: str

    def message(self) -> str:
        return (
            f"Container '{self.container_id}' is missing property '{self.property_}'. "
            f"This property will be skipped for view '{self.view_id}'."
        )

    def dump(self) -> dict[str, str]:
        return {"view_id": self.view_id, "property": self.property_, "container_id": self.container_id}


@dataclass(frozen=True)
class UnknownPropertyTypeWarning(ModelImportWarning):
    description = "Unknown property type. This will be ignored in the imports."
    fix = "Set to a supported property type."
    view_id: str
    property_id: str
    property_type: str

    def message(self) -> str:
        return (
            f"Unknown property type '{self.property_type}' for property '{self.property_id}' "
            f"of view '{self.view_id}'. This will be ignored in the imports."
        )

    def dump(self) -> dict[str, str]:
        return {"view_id": self.view_id, "property_id": self.property_id, "property_type": self.property_type}


@dataclass(frozen=True)
class UnknownContainerConstraintWarning(ModelImportWarning):
    description = "Unknown container constraint. This will be ignored in the imports."
    fix = "Set to a supported container constraint."
    container_id: str
    property_id: str
    constraint: str

    def message(self) -> str:
        return (
            f"Unknown container constraint '{self.constraint}' for property '{self.property_id}' of container "
            f"'{self.container_id}'. This will be ignored in the imports."
        )

    def dump(self) -> dict[str, str]:
        return {"container_id": self.container_id, "property_id": self.property_id, "constraint": self.constraint}


@dataclass(frozen=True)
class FailedToInferValueTypeWarning(ModelImportWarning):
    description = "Failed to infer value type. This will be ignored in the imports."
    fix = "Set to a supported value type."
    view_id: str
    property_id: str

    def message(self) -> str:
        return (
            f"Failed to infer value type for property '{self.property_id}' of view '{self.view_id}'. "
            f"This will be ignored in the imports."
        )

    def dump(self) -> dict[str, str]:
        return {"view_id": self.view_id, "property_id": self.property_id}


@dataclass(frozen=True)
class APIWarning(ModelImportWarning):
    description = "An error was raised."
    fix = "No fix is available."

    error_message: str

    def message(self) -> str:
        return self.error_message

    def dump(self) -> dict[str, str]:
        return {"error_message": self.error_message}


@dataclass(frozen=True)
class FailedImportWarning(ModelImportWarning):
    description = "Failed to import part of the model."
    fix = "No fix is available."

    identifier: set[str]

    def message(self) -> str:
        return f"Failed to import: {self.identifier}. This will be skipped."

    def dump(self) -> dict[str, Any]:
        return {"identifier": list(self.identifier)}


@dataclass(frozen=True)
class MoreThanOneNonAlphanumericCharacterWarning(ModelImportWarning):
    description = """This warning is raised when doing regex validation of strings which either represent class ids,
    property ids, prefix, data model name, that contain more than one non-alphanumeric character,
    such as for example '_' or '-'."""

    field_name: str
    value: str

    def message(self) -> str:
        return f"Field {self.field_name} with value {self.value} contains more than one non-alphanumeric character!"

    def dump(self) -> dict[str, str]:
        return {"field_name": self.field_name, "value": self.value}


@dataclass(frozen=True)
class ModelImportError(NeatValidationError, ABC):
    description = "An error was raised during importing."
    fix = "No fix is available."


@dataclass(frozen=True)
class NoDataModelError(ModelImportError):
    description = "No data model found.."
    fix = "Check if the data model exists in the source."

    error_message: str

    def message(self) -> str:
        return self.error_message

    def dump(self) -> dict[str, str]:
        return {"error_message": self.error_message}


@dataclass(frozen=True)
class APIError(ModelImportError):
    description = "An error was raised during importing."
    fix = "No fix is available."

    error_message: str

    def message(self) -> str:
        return self.error_message

    def dump(self) -> dict[str, str]:
        return {"error_message": self.error_message}


@dataclass(frozen=True)
class InvalidComponentError(ModelImportError, ABC):
    description = "This is a base class for all errors related invalid component definitions"
    fix = "Check if the component is defined in the DTDL file."

    component_type: str
    instance_name: str | None = None
    instance_id: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["component_type"] = self.component_type
        output["instance_name"] = self.instance_name
        output["instance_id"] = self.instance_id
        return output


@dataclass(frozen=True)
class MissingParentDefinitionError(InvalidComponentError):
    description = "The parent component is missing"
    fix = "Check if the parent component is defined in the DTDL file."

    def message(self) -> str:
        if self.instance_name:
            return f"Parent component '{self.component_type}' with name '{self.instance_name}' is missing."
        else:
            return f"Parent component '{self.component_type}' is missing."


@dataclass(frozen=True)
class MissingIdentifierError(InvalidComponentError):
    description = "The identifier is missing"
    fix = "Check if the identifier is defined in the DTDL file."

    component_type: str
    instance_name: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["component_type"] = self.component_type
        output["instance_name"] = self.instance_name
        return output

    def message(self) -> str:
        if self.instance_name:
            return f"Identifier for component '{self.component_type}' with name '{self.instance_name}' is missing."
        else:
            return f"Identifier for component '{self.component_type}' is missing."


@dataclass(frozen=True)
class UnsupportedPropertyTypeError(ModelImportError):
    description = "The property type is not supported"
    fix = "Check if the property type is defined in the DTDL file."
    component_type: str
    property_name: str
    property_type: str
    instance_name: str | None = None
    instance_id: str | None = None

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["component_type"] = self.component_type
        output["property_name"] = self.property_name
        output["property_type"] = self.property_type
        output["instance_name"] = self.instance_name
        output["instance_id"] = self.instance_id
        return output

    def message(self) -> str:
        if self.instance_name:
            return (
                f"Property '{self.property_name}' of type '{self.property_type}' "
                f"of instance '{self.instance_name}' ({self.component_type}) is not supported."
            )
        else:
            return (
                f"Property '{self.property_name}' of type '{self.property_type}' "
                f"of instance '{self.instance_id}' ({self.component_type}) is not supported."
            )



from __future__ import annotations

import logging
from typing import ClassVar

from pydantic import BaseModel, ConfigDict, root_validator, validator
from rdflib import Namespace

from cognite.neat.legacy.rules.models.rules import METADATA_VALUE_MAX_LENGTH


class AssetClassMapping(BaseModel):
    external_id: str
    name: str
    parent_external_id: str | None = None
    description: str | None = None
    metadata: dict | None = {}

    @root_validator(pre=True)
    def create_metadata(cls, values: dict):
        fields = values.keys()

        # adding metadata key in case if it is missing
        values["metadata"] = {} if "metadata" not in values else values["metadata"]

        for field in fields:
            if field not in ["external_id", "name", "parent_external_id", "data_set_id", "metadata", "description"]:
                values["metadata"][field] = ""
        return values


class AssetTemplate(BaseModel):
    """This class is used to validate, repair and wrangle rdf asset dictionary according to the
    expected format of cognite sdk Asset dataclass."""

    external_id_prefix: str | None = None  # convenience field to add prefix to external_ids
    external_id: str
    name: str | None = None
    parent_external_id: str | None = None
    metadata: dict | None = {}
    description: str | None = None
    data_set_id: int | None = None

    @root_validator(pre=True)
    def preprocess_fields(cls, values: dict):
        fields = values.keys()

        # Adding metadata key in case if it is missing
        values["metadata"] = {} if "metadata" not in values else values["metadata"]

        for field in fields:
            # Enrich: adding any field that is not in the list of expected fields to metadata
            if field not in [
                "external_id",
                "name",
                "parent_external_id",
                "data_set_id",
                "metadata",
                "description",
                "external_id_prefix",
            ]:
                values["metadata"][field] = values[field]

            # Repair: in case if name/description is list instead of single value list elements are joined
            elif field in ["name", "description"] and isinstance(values[field], list):
                msg = f"{values['type']} instance {values['identifier']} property {field} "
                msg += f"has multiple values {values[field]}, "
                msg += f"these values will be joined in a single string: {', '.join(values[field])}"
                logging.info(msg)
                values[field] = ", ".join(sorted(values[field]))[: METADATA_VALUE_MAX_LENGTH - 1]

            # Repair: in case if external_id or parent_external_id are lists, we take the first value
            elif field in ["external_id", "parent_external_id"] and isinstance(values[field], list):
                msg = f"{values['type']} instance {values['identifier']} property {field} "
                msg += f"has multiple values {values[field]}, "
                msg += f"only the first one will be used: {values[field][0]}"
                logging.info(msg)
                values[field] = values[field][0]

        # Setting asset to be by default active
        values["metadata"]["active"] = "true"

        # Handling case when the external_id is not provided by defaulting to the original identifier
        # The original identifier probably has its namespace removed
        if "external_id" not in fields and "identifier" in fields:
            values["external_id"] = values["identifier"]

        return values

    @validator("metadata")
    def to_list_if_comma(cls, value):
        for key, v in value.items():
            if isinstance(v, list):
                value[key] = ", ".join(sorted(v))[: METADATA_VALUE_MAX_LENGTH - 1]
        return value

    @validator("metadata")
    def to_str(cls, value):
        for key, v in value.items():
            value[key] = str(v)
        return value

    @validator("external_id", always=True)
    def add_prefix_to_external_id(cls, value, values):
        if values["external_id_prefix"]:
            return values["external_id_prefix"] + value
        else:
            return value

    @validator("parent_external_id")
    def add_prefix_to_parent_external_id(cls, value, values):
        if values["external_id_prefix"]:
            return values["external_id_prefix"] + value
        else:
            return value


class RelationshipDefinition(BaseModel):
    source_class: str
    target_class: str
    property_: str
    labels: list[str] | None = None
    target_type: str = "Asset"
    source_type: str = "Asset"
    relationship_external_id_rule: str | None = None


class RelationshipDefinitions(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False
    )

    data_set_id: int
    prefix: str
    namespace: Namespace
    relationships: dict[str, RelationshipDefinition]



"""Methods to transform Domain Knowledge Graph to App Knowledge Graph"""

import logging
import time
import traceback
from typing import Any

import pandas as pd
from cognite.client import CogniteClient
from prometheus_client import Gauge
from pydantic import BaseModel
from rdflib import RDF, Graph
from rdflib.term import Literal, Node

from cognite.neat.legacy.graph.exceptions import NamespaceRequired
from cognite.neat.legacy.graph.transformations.query_generator.sparql import build_sparql_query
from cognite.neat.legacy.rules.models._base import EntityTypes
from cognite.neat.legacy.rules.models.rdfpath import (
    AllProperties,
    AllReferences,
    Query,
    RawLookup,
    Traversal,
    parse_rule,
)
from cognite.neat.legacy.rules.models.rules import Rules
from cognite.neat.utils.utils import remove_namespace

prom_total_proc_rules_g = Gauge("neat_total_processed_rules", "Number of processed rules", ["state"])
rules_processing_timing_metric = Gauge(
    "neat_rules_processing_timing", "Transformation rules processing timing metrics", ["aggregation"]
)


COMMIT_BATCH_SIZE = 10000


class RuleProcessingReportRec(BaseModel):
    """Report record for rule processing"""

    row_id: str | None = None
    rule_name: str | None = None
    rule_type: str | None = None
    rule_expression: Any | None = None
    status: str | None = None
    error_message: str | None = None
    elapsed_time: float = 0
    rows_in_response: int = 0


class RuleProcessingReport(BaseModel):
    """Report for rule processing"""

    total_rules: int = 0
    total_success: int = 0
    total_success_no_results: int = 0
    total_failed: int = 0
    records: list[RuleProcessingReportRec] = []
    elapsed_time: float = 0


def source2solution_graph(
    source_knowledge_graph: Graph,
    transformation_rules: Rules,
    solution_knowledge_graph: Graph | None = None,
    client: CogniteClient | None = None,
    cdf_lookup_database: str | None = None,
    extra_triples: list[tuple[Node, Node, Node]] | None = None,
    stop_on_exception: bool = False,
    missing_raw_lookup_value: str = "NaN",
    processing_report: RuleProcessingReport | None = None,
) -> Graph:
    """Transforms solution knowledge graph based on Domain Knowledge Graph

    Args:
        source_knowledge_graph: Domain Knowledge Graph which represents the source graph being
                                transformed to app/solution specific graph
        transformation_rules: Transformation rules holding data model definition and rules to
                              transform source/domain graph to app/solution specific graph
        solution_knowledge_graph: Graph to store app/solution specific graph.
                            Defaults to None (i.e., empty graph).
        client: CogniteClient. Defaults to None.
        cdf_lookup_database: CDF RAW database name to use for `rawlookup` rules. Defaults to None.
        extra_triples: Additional triples to add to app/solution knowledge graph. Defaults to None.
        stop_on_exception: To stop on exception. Defaults to False.
        missing_raw_lookup_value: If no value is find for `rawlookup` default value to use. Defaults to "NaN".
        processing_report: Processing report to store results to. Defaults to None.

    Returns:
        Transformed knowledge graph based on transformation rules
    """

    # TODO: This is to be improved and slowly sunset domain2app_knowledge_graph

    return domain2app_knowledge_graph(
        domain_knowledge_graph=source_knowledge_graph,
        transformation_rules=transformation_rules,
        app_instance_graph=solution_knowledge_graph,
        client=client,
        cdf_lookup_database=cdf_lookup_database,
        extra_triples=extra_triples,
        stop_on_exception=stop_on_exception,
        missing_raw_lookup_value=missing_raw_lookup_value,
        processing_report=processing_report,
    )


def domain2app_knowledge_graph(
    domain_knowledge_graph: Graph,
    transformation_rules: Rules,
    app_instance_graph: Graph | None = None,
    client: CogniteClient | None = None,
    cdf_lookup_database: str | None = None,
    extra_triples: list[tuple[Node, Node, Node]] | None = None,
    stop_on_exception: bool = False,
    missing_raw_lookup_value: str = "NaN",
    processing_report: RuleProcessingReport | None = None,
) -> Graph:
    """Generates application/solution specific knowledge graph based on Domain Knowledge Graph

    Args:
        domain_knowledge_graph: Domain Knowledge Graph which represent the source graph being
                                transformed to app/solution specific graph
        transformation_rules: Transformation rules holding data model definition and rules
                              to transform source/domain graph to app/solution specific graph
        app_instance_graph: Graph to store app/solution specific graph. Defaults to None (i.e., empty graph).
        client: CogniteClient. Defaults to None.
        cdf_lookup_database: CDF RAW database name to use for `rawlookup` rules. Defaults to None.
        extra_triples: Additional triples to add to app/solution knowledge graph. Defaults to None.
        stop_on_exception: To stop on exception. Defaults to False.
        missing_raw_lookup_value: If no value is find for `rawlookup` default value to use. Defaults to "NaN".
        processing_report: Processing report to store results to. Defaults to None.

    Returns:
        Transformed knowledge graph based on transformation rules
    """
    if transformation_rules.metadata.namespace is None:
        raise NamespaceRequired("Transform domain to app knowledge graph")
    rule_namespace = transformation_rules.metadata.namespace

    if app_instance_graph is None:
        app_instance_graph = Graph()
        # Bind App namespace and prefix
        app_instance_graph.bind(transformation_rules.metadata.prefix, rule_namespace)
        # Bind other prefixes and namespaces
        for prefix, namespace in transformation_rules.prefixes.items():
            app_instance_graph.bind(prefix, namespace)

    tables_by_name = {}
    if cdf_lookup_database and client:
        for table_name in transformation_rules.raw_tables:
            logging.debug(f"Loading {table_name} table from database {cdf_lookup_database}")
            table = client.raw.rows.retrieve_dataframe(cdf_lookup_database, table_name, limit=-1)
            tables_by_name[table_name] = table

    # Add references with their type first
    types = []
    success = 0
    success_no_results = 0
    failed = 0
    commit_counter = 0
    timing_traces = []
    prom_total_proc_rules_g.labels(state="success_no_results").set(0)
    prom_total_proc_rules_g.labels(state="success").set(0)
    prom_total_proc_rules_g.labels(state="failed").set(0)

    def check_commit(force_commit: bool = False):
        """'Commit nodes to the graph if batch counter is reached or if force_commit is True"""

        if force_commit:
            logging.debug("Committing nodes")
            app_instance_graph.commit()
            logging.debug("Nodes committed")
            return
        nonlocal commit_counter
        commit_counter += 1
        if commit_counter >= COMMIT_BATCH_SIZE:
            logging.info(f"Committing {COMMIT_BATCH_SIZE} nodes")
            app_instance_graph.commit()
            logging.info(f" {COMMIT_BATCH_SIZE} nodes committed")
            commit_counter = 0

    proc_start_time = time.perf_counter()
    for sheet_row, rule_definition in transformation_rules.properties.items():
        if not rule_definition.rule or rule_definition.skip_rule:
            continue
        msg = f"Processing {sheet_row}: class <{rule_definition.class_id}> "
        msg += f"property <{rule_definition.property_name}> rule <{rule_definition.rule}>"

        processing_report_rec = RuleProcessingReportRec(
            row_id=sheet_row,
            rule_name=f"{rule_definition.class_id}_{rule_definition.property_name}",
            rule_type=rule_definition.rule_type,
            rule_expression=rule_definition.rule,
        )
        logging.info(msg)
        try:
            start_time = time.perf_counter()
            # Parse rule:
            rule = parse_rule(rule_definition.rule, rule_definition.rule_type)  # type: ignore[arg-type]

            # Build SPARQL if needed:
            if isinstance(rule.traversal, Query) and rule_definition.rule_type == "sparql":
                query = rule.traversal.query
            elif isinstance(rule.traversal, Traversal | str):
                query = build_sparql_query(domain_knowledge_graph, rule.traversal, transformation_rules.prefixes)
            else:
                raise ValueError(f"Unknown traversal type {type(rule.traversal)}")
            logging.debug(f"Query: {query}")

            if query_results := list(domain_knowledge_graph.query(query)):
                # Generate URI for class and property in target namespace
                class_URI = rule_namespace[rule_definition.class_id]
                property_URI = rule_namespace[rule_definition.property_name]  # type: ignore[index]

                # Turn query results into dataframe
                instance_df = pd.DataFrame(
                    query_results, columns=[EntityTypes.subject, EntityTypes.predicate, EntityTypes.object]
                )

                # If we are not grabbing all properties for class instances
                # then we are able to replace source property URI with target property URI
                # otherwise we should keep source property URI
                if not isinstance(rule.traversal, AllProperties):
                    instance_df[EntityTypes.predicate] = property_URI

                # If we are storing object from the source graph as literal value(propety type being Datatype Property)
                # in the target graph then we should remove namespace from the object URI and store it as literal
                if isinstance(rule.traversal, AllReferences) and rule_definition.property_type == "DatatypeProperty":
                    instance_df[EntityTypes.object] = instance_df[EntityTypes.object].apply(
                        lambda x: Literal(remove_namespace(x))
                    )

                if isinstance(rule, RawLookup):
                    lookup_map = tables_by_name[rule.table.name].set_index(rule.table.key)[rule.table.value].to_dict()

                    def lookup(
                        literal: Literal, lookup_table=lookup_map, missing_raw_lookup_value=missing_raw_lookup_value
                    ):
                        if new_value := lookup_table.get(literal.value):
                            return Literal(new_value, literal.language, literal.datatype, bool(literal.normalize))
                        elif missing_raw_lookup_value:
                            return Literal(
                                missing_raw_lookup_value, literal.language, literal.datatype, bool(literal.normalize)
                            )
                        else:
                            return literal

                    instance_df[EntityTypes.object] = instance_df[EntityTypes.object].apply(lookup)

                # Add instances
                for _, triple in instance_df.iterrows():
                    app_instance_graph.add(triple.values)  # type: ignore[arg-type]
                    check_commit()
                # Setting instances type and merging them with df containing instance - type relations
                instance_df[EntityTypes.predicate] = RDF.type
                instance_df[EntityTypes.object] = class_URI
                types.append(instance_df)
                success += 1
                prom_total_proc_rules_g.labels(state="success").inc()
                elapsed_time = time.perf_counter() - start_time
                timing_traces.append(elapsed_time)
                processing_report_rec.elapsed_time = elapsed_time
                processing_report_rec.status = "success"
                processing_report_rec.rows_in_response = len(instance_df)
            else:
                success_no_results += 1
                prom_total_proc_rules_g.labels(state="success_no_results").inc()
                elapsed_time = time.perf_counter() - start_time
                timing_traces.append(elapsed_time)
                processing_report_rec.elapsed_time = elapsed_time
                processing_report_rec.status = "success_no_results"

        except Exception as e:
            failed += 1
            elapsed_time = time.perf_counter() - start_time
            processing_report_rec.elapsed_time = elapsed_time
            processing_report_rec.status = "failed"
            processing_report_rec.error_message = str(e)
            prom_total_proc_rules_g.labels(state="failed").inc()
            logging.error(
                f" Error while processing rule {rule_definition.rule} for class {rule_definition.class_id} \
                and property {rule_definition.property_name}"
            )
            logging.error(traceback.format_exc())
            if stop_on_exception:
                raise e

        if processing_report:
            processing_report.records.append(processing_report_rec)

    if processing_report:
        processing_report.total_rules = len(transformation_rules.properties)
        processing_report.total_success = success
        processing_report.total_success_no_results = success_no_results
        processing_report.total_failed = failed
        processing_report.elapsed_time = time.perf_counter() - proc_start_time

    if timing_traces:
        df = pd.Series(timing_traces)
        rules_processing_timing_metric.labels(aggregation="sum").set(df.sum())
        rules_processing_timing_metric.labels(aggregation="std_div").set(df.std())
        rules_processing_timing_metric.labels(aggregation="min").set(df.min())
        rules_processing_timing_metric.labels(aggregation="max").set(df.max())
        rules_processing_timing_metric.labels(aggregation="mean").set(df.mean())

    type_df = pd.concat(types).drop_duplicates(EntityTypes.subject).reset_index(drop=True)

    # Add instance - RDF Type relations
    for _, triple in type_df.iterrows():  # type: ignore[assignment]
        app_instance_graph.add(triple.values)  # type: ignore[arg-type]
        check_commit()

    for i, triple in enumerate(extra_triples or []):  # type: ignore[assignment]
        try:
            app_instance_graph.add(triple)  # type: ignore[arg-type]
            check_commit()
        except ValueError as e:
            raise ValueError(f"Triple {i} in extra_triples is not correct and cannot be added!") from e

    check_commit(force_commit=True)
    return app_instance_graph
